use std::{
    collections::{hash_map::Entry, HashMap, HashSet},
    num::NonZero,
    sync::Arc,
};

use winit::event_loop::EventLoopProxy;

use crate::{emulator::EmulatorClient, window::WindowManager};

pub struct Application {
    client: EmulatorClient,
    proxy: EventLoopProxy<UserEvent>,
    state: Option<AppState>,
}

impl Application {
    pub fn new(client: EmulatorClient, proxy: EventLoopProxy<UserEvent>) -> Self {
        Self {
            client,
            proxy,
            state: None,
        }
    }
}

impl winit::application::ApplicationHandler<UserEvent> for Application {
    fn resumed(&mut self, event_loop: &winit::event_loop::ActiveEventLoop) {
        self.state = Some(AppState::new(
            self.client.clone(),
            self.proxy.clone(),
            event_loop,
        ));
    }

    fn window_event(
        &mut self,
        event_loop: &winit::event_loop::ActiveEventLoop,
        window_id: winit::window::WindowId,
        event: winit::event::WindowEvent,
    ) {
        self.state
            .as_mut()
            .unwrap()
            .window_event(event_loop, window_id, event);
    }

    fn device_event(
        &mut self,
        _event_loop: &winit::event_loop::ActiveEventLoop,
        _device_id: winit::event::DeviceId,
        event: winit::event::DeviceEvent,
    ) {
        self.state.as_mut().unwrap().device_event(event)
    }

    fn user_event(&mut self, _event_loop: &winit::event_loop::ActiveEventLoop, event: UserEvent) {
        self.state.as_mut().unwrap().user_event(event);
    }
}

struct AppState {
    painter: egui_wgpu::winit::Painter,
    ctx: egui::Context,
    viewports: HashMap<egui::ViewportId, ViewportState>,
    viewports_by_window: HashMap<winit::window::WindowId, egui::ViewportId>,
    focused: Option<egui::ViewportId>,
    screen: WindowManager,
}
impl AppState {
    fn new(
        client: EmulatorClient,
        proxy: EventLoopProxy<UserEvent>,
        event_loop: &winit::event_loop::ActiveEventLoop,
    ) -> Self {
        let mut painter = egui_wgpu::winit::Painter::new(
            egui_wgpu::WgpuConfiguration::default(),
            1,
            None,
            false,
            true,
        );
        let ctx = egui::Context::default();
        ctx.style_mut(|s| {
            s.wrap_mode = Some(egui::TextWrapMode::Extend);
            s.visuals.menu_rounding = Default::default();
        });
        ctx.set_embed_viewports(false);
        {
            let proxy = proxy.clone();
            ctx.set_request_repaint_callback(move |info| {
                proxy.send_event(UserEvent::RepaintRequested(info)).unwrap();
            });
        }

        let mut screen = WindowManager::new(client, proxy);
        let root_viewport = ViewportState::new(
            egui::ViewportId::ROOT,
            &ctx,
            event_loop,
            screen.initial_viewport(),
            &mut painter,
        );
        screen.init_renderer(painter.render_state().as_ref().unwrap());

        let mut viewports_by_window = HashMap::new();
        viewports_by_window.insert(root_viewport.window.id(), egui::ViewportId::ROOT);
        let mut viewports = HashMap::new();
        viewports.insert(egui::ViewportId::ROOT, root_viewport);

        Self {
            painter,
            ctx,
            viewports,
            viewports_by_window,
            focused: Some(egui::ViewportId::ROOT),
            screen,
        }
    }

    fn window_event(
        &mut self,
        event_loop: &winit::event_loop::ActiveEventLoop,
        window_id: winit::window::WindowId,
        event: winit::event::WindowEvent,
    ) {
        let Some(&viewport_id) = self.viewports_by_window.get(&window_id) else {
            return;
        };
        let Some(viewport) = self.viewports.get_mut(&viewport_id) else {
            panic!("Unrecognized viewport");
        };
        viewport.on_window_event(&event);

        match event {
            winit::event::WindowEvent::KeyboardInput { event, .. } => {
                self.screen.handle_key_event(event);
            }
            winit::event::WindowEvent::RedrawRequested => {
                pollster::block_on(
                    self.painter
                        .set_window(viewport_id, Some(viewport.window.clone())),
                )
                .unwrap();
                let cb = viewport.viewport_ui_cb.clone();
                let mut input = viewport.state.take_egui_input(&viewport.window);
                input.viewports = self
                    .viewports
                    .iter()
                    .map(|(k, v)| (*k, v.info.clone()))
                    .collect();
                let output = self.ctx.run(input, |ctx| {
                    if let Some(cb) = cb.as_deref() {
                        cb(ctx)
                    } else {
                        self.screen.show(ctx)
                    }
                });
                let clipped_primitives =
                    self.ctx.tessellate(output.shapes, output.pixels_per_point);
                self.painter.paint_and_update_textures(
                    viewport_id,
                    output.pixels_per_point,
                    [0.0, 0.0, 0.0, 0.0],
                    &clipped_primitives,
                    &output.textures_delta,
                    false,
                );

                let mut live_viewports = HashSet::default();
                for (id, output) in output.viewport_output {
                    live_viewports.insert(id);
                    let viewport = match self.viewports.entry(id) {
                        Entry::Occupied(e) => e.into_mut(),
                        Entry::Vacant(e) => {
                            let mut v = ViewportState::new(
                                id,
                                &self.ctx,
                                event_loop,
                                output.builder,
                                &mut self.painter,
                            );
                            v.viewport_ui_cb = output.viewport_ui_cb;
                            self.viewports_by_window.insert(v.window.id(), id);
                            e.insert(v)
                        }
                    };
                    egui_winit::process_viewport_commands(
                        &self.ctx,
                        &mut viewport.info,
                        output.commands,
                        &viewport.window,
                        &mut HashSet::default(),
                    );
                    if viewport.info.close_requested() {
                        live_viewports.remove(&id);
                    }
                }
                self.viewports.retain(|k, v| {
                    if live_viewports.contains(k) {
                        return true;
                    }
                    self.viewports_by_window.remove(&v.window.id());
                    false
                });
                self.painter.gc_viewports(&live_viewports);
                if !self.viewports.contains_key(&egui::ViewportId::ROOT) {
                    event_loop.exit();
                }
            }
            winit::event::WindowEvent::Resized(size) => {
                let (Some(width), Some(height)) =
                    (NonZero::new(size.width), NonZero::new(size.height))
                else {
                    return;
                };
                self.painter.on_window_resized(viewport_id, width, height);
            }
            winit::event::WindowEvent::Focused(new_focused) => {
                self.focused = new_focused.then_some(viewport_id);
            }
            winit::event::WindowEvent::CloseRequested => {
                if viewport_id == egui::ViewportId::ROOT {
                    event_loop.exit();
                } else if let Some(viewport) = self.viewports.get_mut(&viewport_id) {
                    viewport.info.events.push(egui::ViewportEvent::Close);
                    self.ctx.request_repaint_of(viewport_id);
                    self.ctx.request_repaint_of(egui::ViewportId::ROOT);
                }
            }
            _ => {}
        }
    }

    fn device_event(&mut self, event: winit::event::DeviceEvent) {
        if let winit::event::DeviceEvent::MouseMotion { delta } = event {
            let Some(viewport) = self
                .focused
                .as_ref()
                .and_then(|id| self.viewports.get_mut(id))
            else {
                return;
            };
            viewport.state.on_mouse_motion(delta);
        }
    }

    fn user_event(&mut self, event: UserEvent) {
        match event {
            UserEvent::GamepadEvent(event) => self.screen.handle_gamepad_event(event),
            UserEvent::RepaintRequested(info) => {
                let Some(viewport) = self.viewports.get(&info.viewport_id) else {
                    return;
                };
                viewport.window.request_redraw();
            }
        }
    }
}

struct ViewportState {
    info: egui::ViewportInfo,
    state: egui_winit::State,
    viewport_ui_cb: Option<Arc<egui::DeferredViewportUiCallback>>,
    window: Arc<winit::window::Window>,
}

impl ViewportState {
    fn new(
        id: egui::ViewportId,
        ctx: &egui::Context,
        event_loop: &winit::event_loop::ActiveEventLoop,
        viewport: egui::ViewportBuilder,
        painter: &mut egui_wgpu::winit::Painter,
    ) -> Self {
        let mut info = egui::ViewportInfo::default();
        let window = Arc::new(egui_winit::create_window(ctx, event_loop, &viewport).unwrap());
        egui_winit::update_viewport_info(&mut info, ctx, &window, true);

        pollster::block_on(painter.set_window(id, Some(window.clone()))).unwrap();
        let state = egui_winit::State::new(
            ctx.clone(),
            id,
            event_loop,
            Some(window.scale_factor() as f32),
            event_loop.system_theme(),
            painter.max_texture_side(),
        );
        Self {
            info,
            state,
            viewport_ui_cb: None,
            window,
        }
    }

    fn on_window_event(&mut self, event: &winit::event::WindowEvent) {
        let response = self.state.on_window_event(&self.window, event);
        if response.repaint {
            self.window.request_redraw();
        }
        egui_winit::update_viewport_info(
            &mut self.info,
            self.state.egui_ctx(),
            &self.window,
            false,
        );
    }
}

#[derive(Debug)]
pub enum UserEvent {
    GamepadEvent(gilrs::Event),
    RepaintRequested(egui::RequestRepaintInfo),
}
